#include <pch.h>

class Window {
public:
    static Window& Get() {
        static Window window;
        return window;
    }

    Window(Window const&) = delete;
    void operator=(Window const&) = delete;

    void SetTitle(const char* title) {
        glfwSetWindowTitle(handle, title);
    }

    bool ShouldClose() {
        return glfwWindowShouldClose(handle);
    }

    bool GetKey(int key) {
        return glfwGetKey(handle, key);
    }

    void SwapBuffers() {
        glfwSwapBuffers(handle);
    }

    GLFWwindow* GetHandle() {
        return handle;
    }

private:
    Window() {
        handle = glfwCreateWindow(800, 600, "Ember", nullptr, nullptr);
        glfwMakeContextCurrent(handle);
    }

    ~Window() {
        glfwDestroyWindow(handle);
    }

    GLFWwindow* handle;
};

class Input {
public:
    static Input& Get() {
        static Input input;
        return input;
    }

    void Update() {
        glfwPollEvents();

        for(int key = Keys::SPACE; key < Keys::COUNT; key++) {
            previous_keys[key] = current_keys[key];
            current_keys[key] = IsKeyDown(key);
        }
    }

    bool IsKeyDown(int key) {
        return Window::Get().GetKey(key);
    }

    bool IsKeyUp(int key) {
        return !IsKeyDown(key);
    }

    bool IsKeyPressed(int key) {
        return current_keys[key] && !previous_keys[key];
    }

    bool IsKeyReleased(int key) {
        return !current_keys[key] && previous_keys[key];
    }

private:
    Input() = default;

    bool current_keys[Keys::COUNT];
    bool previous_keys[Keys::COUNT];
};

struct Type {
    GLenum gl_type;
    size_t size;
};

namespace Types {
    const Type FLOAT = { GL_FLOAT, sizeof(float) };
};

struct VertexBufferAttribute {
    Type type;
    GLint count;
};

class VertexBuffer {
public:
    VertexBuffer() {
        glCreateBuffers(1, &id);
    }

    void SetData(void* data, GLsizeiptr size, GLenum usage) {
        data_size = size;
        glNamedBufferData(id, data_size, data, usage);
    }

    void AddAttribute(Type type, GLint count) {
        attributes.emplace_back(type, count);
    }

    GLuint id;
    GLsizeiptr data_size;
    std::vector<VertexBufferAttribute> attributes;
};

class VertexArray {
public:
    VertexArray() {
        glCreateVertexArrays(1, &id);
    }

    void AddVertexBuffer(VertexBuffer vertex_buffer) {
        size_t stride = 0;
        for(const VertexBufferAttribute& attrib : vertex_buffer.attributes) {
            stride += attrib.type.size * attrib.count;
        }

        glVertexArrayVertexBuffer(id, buffer_count, vertex_buffer.id, 0, stride);

        for(size_t i = 0; i < vertex_buffer.attributes.size(); i++) {
            const VertexBufferAttribute& attrib = vertex_buffer.attributes[i];

            glEnableVertexArrayAttrib(id, attribute_count);
            glVertexArrayAttribFormat(id, attribute_count, attrib.count, attrib.type.gl_type, GL_FALSE,
                i * attrib.type.size * attrib.count);

            glVertexArrayAttribBinding(id, attribute_count, buffer_count);
            attribute_count++;
        }

        buffer_count++;
    }

    void Draw() {
        glBindVertexArray(id);
        glDrawArrays(GL_TRIANGLES, 0, 3);
    }

private:
    GLuint id;
    GLuint buffer_count = 0;
    GLuint attribute_count = 0;
};

void GLFWCallback(int code, const char* message) {
    std::cerr << std::format("[GLFW Error {}] {}", code, message) << std::endl;
}

class File {
public:
    File(const char* filename) {
        file = fopen(filename, "rwb");
        this->filename = std::string(filename);
    }

    std::string Read() {
        fseek(file, 0, SEEK_END);
        size_t size = ftell(file);
        rewind(file);

        std::string buffer;
        buffer.resize(size);
        fread(&buffer[0], sizeof(char), size, file);

        return buffer;
    }

    ~File() {
        fclose(file);
    }

    std::string filename;

private:
    FILE* file;
};

namespace Log {
    void Log(std::ostream& stream, std::string category, std::string message) {
        stream << std::format("[{}] {}", category, message) << std::endl;
    }

    void Info(std::string message) {
        Log(std::cout, "INFO", message);
    }

    void Warning(std::string message) {
        Log(std::cout, "WARNING", message);
    }

    void Error(std::string message) {
        Log(std::cerr, "ERROR", message);
    }

    void Fatal(std::string message) {
        Log(std::cerr, "FATAL", message);
        exit(EXIT_FAILURE);
    }

    void OpenGLMessage(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length,
        GLchar const* message, void const* user_param) {
        const char* source_string;
        switch(source) {
        case GL_DEBUG_SOURCE_API:
            source_string = "API";
            break;
        case GL_DEBUG_SOURCE_WINDOW_SYSTEM:
            source_string = "WINDOW SYSTEM";
            break;
        case GL_DEBUG_SOURCE_SHADER_COMPILER:
            source_string = "SHADER COMPILER";
            break;
        case GL_DEBUG_SOURCE_THIRD_PARTY:
            source_string = "THIRD PARTY";
            break;
        case GL_DEBUG_SOURCE_APPLICATION:
            source_string = "APPLICATION";
            break;
        case GL_DEBUG_SOURCE_OTHER:
            source_string = "OTHER";
            break;
        }

        const char* type_string;
        switch(type) {
        case GL_DEBUG_TYPE_ERROR:
            type_string = "ERROR";
            break;
        case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
            type_string = "DEPRECATED_BEHAVIOR";
            break;
        case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
            type_string = "UNDEFINED_BEHAVIOR";
            break;
        case GL_DEBUG_TYPE_PORTABILITY:
            type_string = "PORTABILITY";
            break;
        case GL_DEBUG_TYPE_PERFORMANCE:
            type_string = "PERFORMANCE";
            break;
        case GL_DEBUG_TYPE_MARKER:
            type_string = "MARKER";
            break;
        case GL_DEBUG_TYPE_OTHER:
            type_string = "OTHER";
            break;
        }

        const char* severity_string;
        switch(severity) {
        case GL_DEBUG_SEVERITY_NOTIFICATION:
            severity_string = "NOTIFICATION";
            break;
        case GL_DEBUG_SEVERITY_LOW:
            severity_string = "LOW";
            break;
        case GL_DEBUG_SEVERITY_MEDIUM:
            severity_string = "MEDIUM";
            break;
        case GL_DEBUG_SEVERITY_HIGH:
            severity_string = "HIGH";
            break;
        }

        std::cout << std::format("[OpenGL {}] source: {}, severity: {}, type: {}, message: {}", id,
            source_string, severity_string, type_string, message)
                  << std::endl;
    }
}

class Shader {
public:
    Shader(File vs_file, File fs_file) {
        GLuint vertex = glCreateShader(GL_VERTEX_SHADER);
        GLuint fragment = glCreateShader(GL_FRAGMENT_SHADER);

        std::string vs_source = vs_file.Read();
        std::string fs_source = fs_file.Read();

        const char* vs_cstr = vs_source.c_str();
        const char* fs_cstr = fs_source.c_str();

        glShaderSource(vertex, 1, &vs_cstr, nullptr);
        glShaderSource(fragment, 1, &fs_cstr, nullptr);

        Log::Info(std::format("Compiling vertex shader: {}", vs_file.filename));
        glCompileShader(vertex);

        Log::Info(std::format("Compiling fragment shader: {}", fs_file.filename));
        glCompileShader(fragment);

        id = glCreateProgram();
        glAttachShader(id, vertex);
        glAttachShader(id, fragment);
        glLinkProgram(id);
    }

    void Use() {
        glUseProgram(id);
    }

private:
    GLuint id;
};

struct Vertex {
    float pos[3];
    float color[3];
};

int main(int argc, char** argv) {
    glfwSetErrorCallback(GLFWCallback);
    glfwInit();

    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 6);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    Window& window = Window::Get();
    Input& input = Input::Get();

    gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);

    glEnable(GL_DEBUG_OUTPUT);
    glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
    glDebugMessageCallback(Log::OpenGLMessage, nullptr);

    Vertex vertices[] = {
        { { -0.5f, 0.5f, 0.0f }, { 1.0f, 0.0f, 0.0f } },
        { { -0.5f, -0.5f, 0.0f }, { 0.0f, 1.0f, 0.0f } },
        { { 0.5f, -0.5f, 0.0f }, { 0.0f, 0.0f, 1.0f } },
    };

    VertexBuffer buffer;
    buffer.AddAttribute(Types::FLOAT, 3);
    buffer.AddAttribute(Types::FLOAT, 3);
    buffer.SetData(vertices, sizeof(vertices), GL_STATIC_DRAW);

    VertexArray array;
    array.AddVertexBuffer(buffer);

    Shader basic(File("shaders/basic.vert"), File("shaders/basic.frag"));

    while(!window.ShouldClose()) {
        input.Update();

        glClearColor(0.3f, 0.5f, 0.4f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT);

        basic.Use();
        array.Draw();

        window.SwapBuffers();
    }

    return 0;
}
